@@ -4,13 +4,57 @@
 #![deny(clippy::unwrap_used)]
 
 //! A segregated list memory allocator for Vulkan.
+//!
+//! The allocator can pool allocations of a user defined lifetime together to help
+//! reducing the fragmentation.
+//!
+//! ## Example:
+//! ```ignore
+//! #[derive(Debug, Clone, Copy, Hash, Eq, PartialEq)]
+//! enum Lifetime {
+//!     Buffer,
+//!     Image,
+//! }
+//!
+//! impl vk_alloc::Lifetime for Lifetime {}
+//!
+//! Allocator::<Lifetime>::new(
+//!     &instance,
+//!     &physical_device,
+//!     &AllocatorDescriptor {
+//!         ..Default::default()
+//!     },
+//! ).unwrap();
+//!
+//! let allocation = alloc
+//!     .allocate(
+//!         &logical_device,
+//!         &AllocationDescriptor {
+//!             location: MemoryLocation::GpuOnly,
+//!             requirements: vk::MemoryRequirementsBuilder::new()
+//!                 .alignment(512)
+//!                 .size(1024)
+//!                 .memory_type_bits(u32::MAX)
+//!                 .build(),
+//!             lifetime: Lifetime::Buffer,
+//!             is_dedicated: false,
+//!             is_optimal: false,
+//!         },
+//!     )
+//!     .unwrap();
+//!
+//! ```
+//!
+use std::collections::HashMap;
 use std::convert::TryInto;
 use std::ffi::c_void;
+use std::fmt::Debug;
+use std::hash::Hash;
 use std::num::NonZeroUsize;
 use std::ptr;
 
 use erupt::{vk, ExtendableFromMut};
-use parking_lot::Mutex;
+use parking_lot::{Mutex, RwLock};
 #[cfg(feature = "tracing")]
 use tracing1::{debug, info};
 
@@ -23,6 +67,9 @@ type Result<T> = std::result::Result<T, AllocatorError>;
 /// For a minimal bucket size of 256b as log2.
 const MINIMAL_BUCKET_SIZE_LOG2: u32 = 8;
 
+/// The lifetime of an allocation. Used to pool allocations and reduce fragmentation.
+pub trait Lifetime: Debug + Copy + Hash + Eq + PartialEq {}
+
 /// Describes the configuration of an `Allocator`.
 #[derive(Debug, Clone)]
 pub struct AllocatorDescriptor {
@@ -38,17 +85,17 @@ impl Default for AllocatorDescriptor {
 
 /// The general purpose memory allocator. Implemented as a segregated list allocator.
 #[derive(Debug)]
-pub struct Allocator {
+pub struct Allocator<LT: Lifetime> {
     driver_id: vk::DriverId,
     is_integrated: bool,
-    buffer_pools: Vec<Mutex<MemoryPool>>,
-    image_pools: Vec<Mutex<MemoryPool>>,
+    pools: RwLock<HashMap<LT, Vec<Mutex<MemoryPool>>>>,
     block_size: vk::DeviceSize,
-    memory_types_size: u32,
+    memory_types: Vec<vk::MemoryType>,
     memory_properties: vk::PhysicalDeviceMemoryProperties,
+    buffer_image_granularity: u64,
 }
 
-impl Allocator {
+impl<LT: Lifetime> Allocator<LT> {
     /// Creates a new allocator.
     #[cfg_attr(feature = "profiling", profiling::function)]
     pub fn new(
@@ -56,7 +103,8 @@ impl Allocator {
         physical_device: vk::PhysicalDevice,
         descriptor: &AllocatorDescriptor,
     ) -> Result<Self> {
-        let (driver_id, is_integrated) = query_driver(instance, physical_device);
+        let (driver_id, is_integrated, buffer_image_granularity) =
+            query_driver(instance, physical_device);
 
         #[cfg(feature = "tracing")]
         debug!("Driver ID of the physical device: {:?}", driver_id);
@@ -65,49 +113,21 @@ impl Allocator {
             unsafe { instance.get_physical_device_memory_properties(physical_device) };
 
         let memory_types_count: usize = (memory_properties.memory_type_count).try_into()?;
-        let memory_types = &memory_properties.memory_types[..memory_types_count];
+        let memory_types = memory_properties.memory_types[..memory_types_count].to_owned();
 
         #[cfg(feature = "tracing")]
-        print_memory_types(memory_properties, memory_types)?;
-
-        let memory_types_size: u32 = memory_types.len().try_into()?;
+        print_memory_types(memory_properties, &memory_types)?;
 
         let block_size: vk::DeviceSize = (2u64).pow(descriptor.block_size.into()).into();
 
-        let mut buffer_pools = Vec::with_capacity(memory_types_size.try_into()?);
-        for (memory_type, i) in memory_types.iter().zip(0..memory_types_size) {
-            let pool = MemoryPool::new(
-                i,
-                block_size,
-                i,
-                memory_type
-                    .property_flags
-                    .contains(vk::MemoryPropertyFlags::HOST_VISIBLE),
-            )?;
-            buffer_pools.push(Mutex::new(pool));
-        }
-
-        let mut image_pools = Vec::with_capacity(memory_types_size.try_into()?);
-        for (memory_type, i) in memory_types.iter().zip(0..memory_types_size) {
-            let pool = MemoryPool::new(
-                memory_types_size + i,
-                (2u64).pow(descriptor.block_size.into()),
-                i,
-                memory_type
-                    .property_flags
-                    .contains(vk::MemoryPropertyFlags::HOST_VISIBLE),
-            )?;
-            image_pools.push(Mutex::new(pool));
-        }
-
         Ok(Self {
             driver_id,
             is_integrated,
-            buffer_pools,
-            image_pools,
+            pools: RwLock::default(),
             block_size,
-            memory_types_size,
+            memory_types,
             memory_properties,
+            buffer_image_granularity,
         })
     }
 
@@ -117,7 +137,8 @@ impl Allocator {
         device: &erupt::DeviceLoader,
         buffer: vk::Buffer,
         location: MemoryLocation,
-    ) -> Result<Allocation> {
+        lifetime: LT,
+    ) -> Result<Allocation<LT>> {
         let info = vk::BufferMemoryRequirementsInfo2Builder::new().buffer(buffer);
         let mut dedicated_requirements = vk::MemoryDedicatedRequirementsBuilder::new();
         let requirements =
@@ -134,20 +155,23 @@ impl Allocator {
         let alloc_decs = AllocationDescriptor {
             requirements: memory_requirements,
             location,
-            allocation_type: AllocationType::Buffer,
+            lifetime,
             is_dedicated,
+            is_optimal: false,
         };
 
         self.allocate(device, &alloc_decs)
     }
 
-    /// Allocates memory for an image.
+    /// Allocates memory for an image. `is_optimal` must be set true if the image is a optimal image (a regular texture).
     pub fn allocate_memory_for_image(
         &self,
         device: &erupt::DeviceLoader,
         image: vk::Image,
         location: MemoryLocation,
-    ) -> Result<Allocation> {
+        lifetime: LT,
+        is_optimal: bool,
+    ) -> Result<Allocation<LT>> {
         let info = vk::ImageMemoryRequirementsInfo2Builder::new().image(image);
         let mut dedicated_requirements = vk::MemoryDedicatedRequirementsBuilder::new();
         let requirements =
@@ -164,33 +188,21 @@ impl Allocator {
         let alloc_decs = AllocationDescriptor {
             requirements: memory_requirements,
             location,
-            allocation_type: AllocationType::OptimalImage,
+            lifetime,
             is_dedicated,
+            is_optimal,
         };
 
         self.allocate(device, &alloc_decs)
     }
 
     /// Allocates memory on the allocator.
-    //
-    // For each memory type we have two memory pools: For linear resources and for optimal textures.
-    // This removes the need to check for the granularity between them and the idea is, that
-    // buffers/textures have different lifetimes and internal fragmentation is smaller this way.
-    //
-    // Dedicated blocks still exists in their respective pools. They are de-allocated when
-    // they are freed. Normal blocks are not de-allocated.
-    //
-    // Each pool has fixed sized blocks that need to be of power two size. Each block has at
-    // least one chunk.
-    //
-    // Free chunks are saved in a segregated list with buckets of power of two sizes.
-    // The biggest bucket size is the block size.
     #[cfg_attr(feature = "profiling", profiling::function)]
     pub fn allocate(
         &self,
         device: &erupt::DeviceLoader,
-        descriptor: &AllocationDescriptor,
-    ) -> Result<Allocation> {
+        descriptor: &AllocationDescriptor<LT>,
+    ) -> Result<Allocation<LT>> {
         let size = descriptor.requirements.size;
         let alignment = descriptor.requirements.alignment;
 
@@ -209,12 +221,40 @@ impl Allocator {
             descriptor.requirements.memory_type_bits,
         )?;
 
-        let pool = match descriptor.allocation_type {
-            AllocationType::Buffer | AllocationType::LinearImage => {
-                &self.buffer_pools[memory_type_index]
+        let has_key = self.pools.read().contains_key(&descriptor.lifetime);
+        if !has_key {
+            let mut pools = Vec::with_capacity(self.memory_types.len());
+            for (i, memory_type) in self.memory_types.iter().enumerate() {
+                let pool = MemoryPool::new(
+                    self.block_size,
+                    i.try_into()?,
+                    memory_type
+                        .property_flags
+                        .contains(vk::MemoryPropertyFlags::HOST_VISIBLE),
+                )?;
+                pools.push(Mutex::new(pool));
             }
-            AllocationType::OptimalImage => &self.image_pools[memory_type_index],
-        };
+
+            self.pools.write().insert(descriptor.lifetime, pools);
+        }
+
+        let lifetime_pools = self.pools.read();
+
+        let pool = &lifetime_pools
+            .get(&descriptor.lifetime)
+            .ok_or_else(|| {
+                AllocatorError::Internal(format!(
+                    "can't find pool for lifetime {:?}",
+                    descriptor.lifetime
+                ))
+            })?
+            .get(memory_type_index)
+            .ok_or_else(|| {
+                AllocatorError::Internal(format!(
+                    "can't find memory_type {} in pool {:?}",
+                    memory_type_index, descriptor.lifetime
+                ))
+            })?;
 
         if descriptor.is_dedicated || size >= self.block_size {
             #[cfg(feature = "tracing")]
@@ -222,11 +262,19 @@ impl Allocator {
                 "Allocating as dedicated block on memory type {}",
                 memory_type_index
             );
-            pool.lock().allocate_dedicated(device, size)
+            pool.lock()
+                .allocate_dedicated(device, size, descriptor.lifetime)
         } else {
             #[cfg(feature = "tracing")]
             debug!("Sub allocating on memory type {}", memory_type_index);
-            pool.lock().allocate(device, size, alignment)
+            pool.lock().allocate(
+                device,
+                self.buffer_image_granularity,
+                size,
+                alignment,
+                descriptor.lifetime,
+                descriptor.is_optimal,
+            )
         }
     }
 
@@ -321,14 +369,28 @@ impl Allocator {
 
     /// Frees the allocation.
     #[cfg_attr(feature = "profiling", profiling::function)]
-    pub fn deallocate(&self, device: &erupt::DeviceLoader, allocation: &Allocation) -> Result<()> {
-        let memory_pool = if allocation.pool_index > self.memory_types_size {
-            let pool_index: usize = (allocation.pool_index - self.memory_types_size).try_into()?;
-            &self.image_pools[pool_index]
-        } else {
-            let pool_index: usize = allocation.pool_index.try_into()?;
-            &self.buffer_pools[pool_index]
-        };
+    pub fn deallocate(
+        &self,
+        device: &erupt::DeviceLoader,
+        allocation: &Allocation<LT>,
+    ) -> Result<()> {
+        let memory_type_index: usize = allocation.memory_type_index.try_into()?;
+        let pools = &self.pools.read();
+        let memory_pool = &pools
+            .get(&allocation.lifetime)
+            .ok_or_else(|| {
+                AllocatorError::Internal(format!(
+                    "can't find pool for lifetime {:?}",
+                    allocation.lifetime
+                ))
+            })?
+            .get(memory_type_index)
+            .ok_or_else(|| {
+                AllocatorError::Internal(format!(
+                    "can't find memory_type {} in pool {:?}",
+                    memory_type_index, allocation.lifetime
+                ))
+            })?;
 
         if let Some(chunk_key) = allocation.chunk_key {
             #[cfg(feature = "tracing")]
@@ -355,164 +417,186 @@ impl Allocator {
     /// Releases all memory blocks back to the system. Should be called before drop.
     #[cfg_attr(feature = "profiling", profiling::function)]
     pub fn cleanup(&mut self, device: &erupt::DeviceLoader) {
-        self.buffer_pools.drain(..).for_each(|pool| {
-            pool.lock().blocks.iter_mut().for_each(|block| {
-                if let Some(block) = block {
-                    block.destroy(&device)
-                }
-            })
-        });
-        self.image_pools.drain(..).for_each(|pool| {
-            pool.lock().blocks.iter_mut().for_each(|block| {
-                if let Some(block) = block {
-                    block.destroy(&device)
-                }
-            })
-        });
+        for (_, mut lifetime_pools) in self.pools.write().drain() {
+            lifetime_pools.drain(..).for_each(|pool| {
+                pool.lock().blocks.iter_mut().for_each(|block| {
+                    if let Some(block) = block {
+                        block.destroy(&device)
+                    }
+                })
+            });
+        }
     }
 
     /// Number of allocations.
     pub fn allocation_count(&self) -> usize {
-        let mut buffer_count = 0;
-        for pool in &self.buffer_pools {
-            let pool = pool.lock();
-            for chunk in &pool.chunks {
-                if let Some(chunk) = chunk {
-                    if !chunk.is_free {
-                        buffer_count += 1;
+        let mut count = 0;
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                let pool = pool.lock();
+                for chunk in &pool.chunks {
+                    if let Some(chunk) = chunk {
+                        if chunk.chunk_type != ChunkType::Free {
+                            count += 1;
+                        }
                     }
                 }
-            }
+            });
         }
 
-        let mut image_count = 0;
-        for pool in &self.image_pools {
-            let pool = pool.lock();
-            for chunk in &pool.chunks {
-                if let Some(chunk) = chunk {
-                    if !chunk.is_free {
-                        image_count += 1;
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                let pool = pool.lock();
+                for block in &pool.blocks {
+                    if let Some(block) = block {
+                        if block.is_dedicated {
+                            count += 1;
+                        }
                     }
                 }
-            }
+            });
         }
 
-        let mut dedicated_buffer_count = 0;
-        for pool in &self.buffer_pools {
-            let pool = pool.lock();
-            for block in &pool.blocks {
-                if let Some(block) = block {
-                    if block.is_dedicated {
-                        dedicated_buffer_count += 1;
-                    }
-                }
-            }
-        }
-
-        let mut dedicated_image_count = 0;
-        for pool in &self.image_pools {
-            let pool = pool.lock();
-            for block in &pool.blocks {
-                if let Some(block) = block {
-                    if block.is_dedicated {
-                        dedicated_image_count += 1;
-                    }
-                }
-            }
-        }
-
-        buffer_count + image_count + dedicated_buffer_count + dedicated_image_count
+        count
     }
 
     /// Number of unused ranges between allocations.
     pub fn unused_range_count(&self) -> usize {
-        count_unused_ranges(&self.buffer_pools) + count_unused_ranges(&self.image_pools)
+        let mut unused_count: usize = 0;
+
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                collect_start_chunks(&pool).iter().for_each(|key| {
+                    let mut next_key: NonZeroUsize = *key;
+                    let mut previous_size: vk::DeviceSize = 0;
+                    let mut previous_offset: vk::DeviceSize = 0;
+                    loop {
+                        let pool = pool.lock();
+                        let chunk = pool.chunks[next_key.get()]
+                            .as_ref()
+                            .expect("can't find chunk in chunk list");
+                        if chunk.offset != previous_offset + previous_size {
+                            unused_count += 1;
+                        }
+
+                        if let Some(key) = chunk.next {
+                            next_key = key
+                        } else {
+                            break;
+                        }
+
+                        previous_size = chunk.size;
+                        previous_offset = chunk.offset
+                    }
+                });
+            })
+        }
+
+        unused_count
     }
 
     /// Number of bytes used by the allocations.
     pub fn used_bytes(&self) -> vk::DeviceSize {
-        let mut buffer_bytes = 0;
-        for pool in &self.buffer_pools {
-            let pool = pool.lock();
-            for chunk in &pool.chunks {
-                if let Some(chunk) = chunk {
-                    if !chunk.is_free {
-                        buffer_bytes += chunk.size;
+        let mut bytes = 0;
+
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                let pool = pool.lock();
+                for chunk in &pool.chunks {
+                    if let Some(chunk) = chunk {
+                        if chunk.chunk_type != ChunkType::Free {
+                            bytes += chunk.size;
+                        }
                     }
                 }
-            }
+            });
         }
 
-        let mut image_bytes = 0;
-        for pool in &self.image_pools {
-            let pool = pool.lock();
-            for chunk in &pool.chunks {
-                if let Some(chunk) = chunk {
-                    if !chunk.is_free {
-                        image_bytes += chunk.size;
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                let pool = pool.lock();
+                for block in &pool.blocks {
+                    if let Some(block) = block {
+                        if block.is_dedicated {
+                            bytes += block.size;
+                        }
                     }
                 }
-            }
+            });
         }
 
-        let mut dedicated_buffer_bytes = 0;
-        for pool in &self.buffer_pools {
-            let pool = pool.lock();
-            for block in &pool.blocks {
-                if let Some(block) = block {
-                    if block.is_dedicated {
-                        dedicated_buffer_bytes += block.size;
-                    }
-                }
-            }
-        }
-
-        let mut dedicated_image_bytes = 0;
-        for pool in &self.image_pools {
-            let pool = pool.lock();
-            for block in &pool.blocks {
-                if let Some(block) = block {
-                    if block.is_dedicated {
-                        dedicated_image_bytes += block.size;
-                    }
-                }
-            }
-        }
-
-        buffer_bytes + image_bytes + dedicated_buffer_bytes + dedicated_image_bytes
+        bytes
     }
 
     /// Number of bytes used by the unused ranges between allocations.
     pub fn unused_bytes(&self) -> vk::DeviceSize {
-        count_unused_bytes(&self.buffer_pools) + count_unused_bytes(&self.image_pools)
+        let mut unused_bytes: vk::DeviceSize = 0;
+
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            lifetime_pools.iter().for_each(|pool| {
+                collect_start_chunks(&pool).iter().for_each(|key| {
+                    let mut next_key: NonZeroUsize = *key;
+                    let mut previous_size: vk::DeviceSize = 0;
+                    let mut previous_offset: vk::DeviceSize = 0;
+                    loop {
+                        let pool = pool.lock();
+                        let chunk = pool.chunks[next_key.get()]
+                            .as_ref()
+                            .expect("can't find chunk in chunk list");
+                        if chunk.offset != previous_offset + previous_size {
+                            unused_bytes += chunk.offset - (previous_offset + previous_size);
+                        }
+
+                        if let Some(key) = chunk.next {
+                            next_key = key
+                        } else {
+                            break;
+                        }
+
+                        previous_size = chunk.size;
+                        previous_offset = chunk.offset
+                    }
+                });
+            });
+        }
+
+        unused_bytes
     }
 
     /// Number of allocated Vulkan memory blocks.
     pub fn block_count(&self) -> usize {
-        let buffer_sum: usize = self
-            .buffer_pools
-            .iter()
-            .map(|pool| pool.lock().blocks.len())
-            .sum();
-        let image_sum: usize = self
-            .image_pools
-            .iter()
-            .map(|pool| pool.lock().blocks.len())
-            .sum();
+        let mut count: usize = 0;
 
-        buffer_sum + image_sum
+        for (_, lifetime_pools) in self.pools.read().iter() {
+            count += lifetime_pools
+                .iter()
+                .map(|pool| pool.lock().blocks.len())
+                .sum::<usize>();
+        }
+
+        count
     }
 }
 
-/// Type of the allocation.
 #[derive(Debug, Copy, Clone, Eq, PartialEq)]
-pub enum AllocationType {
-    /// An allocation for a buffer.
-    Buffer,
-    /// An allocation for a regular image.
-    OptimalImage,
-    /// An allocation for a linear image.
-    LinearImage,
+#[repr(u8)]
+enum ChunkType {
+    Free,
+    Linear,
+    Optimal,
+}
+
+impl ChunkType {
+    /// There is an implementation-dependent limit, bufferImageGranularity, which specifies a
+    /// page-like granularity at which linear and non-linear resources must be placed in adjacent
+    /// memory locations to avoid aliasing.
+    fn granularity_conflict(self, other: ChunkType) -> bool {
+        if self == ChunkType::Free || other == ChunkType::Free {
+            return false;
+        }
+
+        self != other
+    }
 }
 
 /// The intended location of the memory.
@@ -528,21 +612,25 @@ pub enum MemoryLocation {
 
 /// The descriptor for an allocation on the allocator.
 #[derive(Clone, Debug)]
-pub struct AllocationDescriptor {
+pub struct AllocationDescriptor<LT: Lifetime> {
     /// Location where the memory allocation should be stored.
     pub location: MemoryLocation,
     /// Vulkan memory requirements for an allocation.
     pub requirements: vk::MemoryRequirements,
-    /// The type of the allocation.
-    pub allocation_type: AllocationType,
+    /// The lifetime of an allocation. Used to pool together resources of the same lifetime.
+    pub lifetime: LT,
     /// If the allocation should be dedicated.
     pub is_dedicated: bool,
+    /// True if the allocation is for a optimal image (regular textures). Buffers and linear
+    /// images need to set this false.
+    pub is_optimal: bool,
 }
 
 /// An allocation of the `Allocator`.
 #[derive(Clone, Debug)]
-pub struct Allocation {
-    pool_index: u32,
+pub struct Allocation<LT: Lifetime> {
+    memory_type_index: u32,
+    lifetime: LT,
     block_key: NonZeroUsize,
     chunk_key: Option<NonZeroUsize>,
     mapped_ptr: Option<std::ptr::NonNull<c_void>>,
@@ -555,11 +643,11 @@ pub struct Allocation {
     pub size: vk::DeviceSize,
 }
 
-unsafe impl Send for Allocation {}
+unsafe impl<LT: Lifetime> Send for Allocation<LT> {}
 
-unsafe impl Sync for Allocation {}
+unsafe impl<LT: Lifetime> Sync for Allocation<LT> {}
 
-impl Allocation {
+impl<LT: Lifetime> Allocation<LT> {
     /// Returns a valid mapped slice if the memory is host visible, otherwise it will return None.
     /// The slice already references the exact memory region of the sub allocation, so no offset needs to be applied.
     pub fn mapped_slice(&self) -> Result<Option<&[u8]>> {
@@ -593,6 +681,7 @@ impl Allocation {
 
 #[derive(Clone, Debug)]
 struct BestFitCandidate {
+    aligned_offset: u64,
     key: NonZeroUsize,
     free_list_index: usize,
     free_size: vk::DeviceSize,
@@ -604,9 +693,8 @@ struct BestFitCandidate {
 /// so that internal memory fragmentation is kept low.
 #[derive(Debug)]
 struct MemoryPool {
-    pool_index: u32,
-    block_size: vk::DeviceSize,
     memory_type_index: u32,
+    block_size: vk::DeviceSize,
     is_mappable: bool,
     blocks: Vec<Option<MemoryBlock>>,
     chunks: Vec<Option<MemoryChunk>>,
@@ -619,12 +707,7 @@ struct MemoryPool {
 }
 
 impl MemoryPool {
-    fn new(
-        pool_index: u32,
-        block_size: vk::DeviceSize,
-        memory_type_index: u32,
-        is_mappable: bool,
-    ) -> Result<Self> {
+    fn new(block_size: vk::DeviceSize, memory_type_index: u32, is_mappable: bool) -> Result<Self> {
         let mut blocks = Vec::with_capacity(128);
         let mut chunks = Vec::with_capacity(128);
 
@@ -650,9 +733,8 @@ impl MemoryPool {
         }
 
         Ok(Self {
-            pool_index,
-            block_size,
             memory_type_index,
+            block_size,
             is_mappable,
             blocks,
             chunks,
@@ -685,11 +767,12 @@ impl MemoryPool {
         }
     }
 
-    fn allocate_dedicated(
+    fn allocate_dedicated<LT: Lifetime>(
         &mut self,
         device: &erupt::DeviceLoader,
         size: vk::DeviceSize,
-    ) -> Result<Allocation> {
+        lifetime: LT,
+    ) -> Result<Allocation<LT>> {
         let block = MemoryBlock::new(device, size, self.memory_type_index, self.is_mappable, true)?;
 
         let device_memory = block.device_memory;
@@ -698,7 +781,8 @@ impl MemoryPool {
         let key = self.add_block(block);
 
         Ok(Allocation {
-            pool_index: self.pool_index,
+            memory_type_index: self.memory_type_index,
+            lifetime,
             block_key: key,
             chunk_key: None,
             device_memory,
@@ -708,17 +792,26 @@ impl MemoryPool {
         })
     }
 
-    fn allocate(
+    fn allocate<LT: Lifetime>(
         &mut self,
         device: &erupt::DeviceLoader,
+        buffer_image_granularity: u64,
         size: vk::DeviceSize,
         alignment: vk::DeviceSize,
-    ) -> Result<Allocation> {
+        lifetime: LT,
+        is_optimal: bool,
+    ) -> Result<Allocation<LT>> {
         let mut bucket_index = calculate_bucket_index(size);
 
         // Make sure that we don't try to allocate a chunk bigger than the block.
         debug_assert!(bucket_index <= self.max_bucket_index);
 
+        let chunk_type = if is_optimal {
+            ChunkType::Optimal
+        } else {
+            ChunkType::Linear
+        };
+
         loop {
             // We couldn't find a suitable empty chunk, so we will allocate a new block.
             if bucket_index > self.max_bucket_index {
@@ -735,15 +828,64 @@ impl MemoryPool {
                 let chunk = &self.chunks[key.get()]
                     .as_ref()
                     .expect("can't find chunk in chunk list");
-                debug_assert!(chunk.is_free);
+                debug_assert!(chunk.chunk_type == ChunkType::Free);
 
                 if chunk.size < size {
                     continue;
                 }
 
-                let offset = align_up(chunk.offset, alignment);
-                let padding = offset - chunk.offset;
-                let aligned_size = (padding + size).into();
+                let mut aligned_offset = 0;
+
+                // We need to handle the granularity between chunks. See "Buffer-Image Granularity"
+                // in the Vulkan specs.
+                if let Some(previous) = chunk.previous {
+                    let previous = self
+                        .chunks
+                        .get(previous.get())
+                        .ok_or_else(|| {
+                            AllocatorError::Internal("can't find previous chunk".into())
+                        })?
+                        .as_ref()
+                        .ok_or_else(|| {
+                            AllocatorError::Internal("previous chunk was empty".into())
+                        })?;
+
+                    aligned_offset = align_up(chunk.offset, alignment);
+
+                    if previous.chunk_type.granularity_conflict(chunk_type)
+                        && is_on_same_page(
+                            previous.offset,
+                            previous.size,
+                            aligned_offset,
+                            buffer_image_granularity,
+                        )
+                    {
+                        aligned_offset = align_up(aligned_offset, buffer_image_granularity);
+                    }
+                }
+
+                if let Some(next) = chunk.next {
+                    let next = self
+                        .chunks
+                        .get(next.get())
+                        .ok_or_else(|| AllocatorError::Internal("can't find next chunk".into()))?
+                        .as_ref()
+                        .ok_or_else(|| AllocatorError::Internal("next chunk was empty".into()))?;
+
+                    if next.chunk_type.granularity_conflict(chunk_type)
+                        && is_on_same_page(
+                            next.offset,
+                            next.size,
+                            aligned_offset,
+                            buffer_image_granularity,
+                        )
+                    {
+                        continue;
+                    }
+                }
+
+                let padding = aligned_offset - chunk.offset;
+                let aligned_size = padding + size;
 
                 // Try to find the best fitting chunk.
                 if chunk.size >= aligned_size {
@@ -757,6 +899,7 @@ impl MemoryPool {
 
                     if free_size < best_fit_size {
                         best_fit_candidate = Some(BestFitCandidate {
+                            aligned_offset,
                             key: *key,
                             free_list_index: index,
                             free_size,
@@ -779,12 +922,9 @@ impl MemoryPool {
                         .expect("can't find candidate in chunk list")
                         .clone();
 
-                    let candidate_chunk_aligned_offset =
-                        align_up(candidate_chunk.offset, alignment);
-                    let candidate_chunk_padding =
-                        candidate_chunk_aligned_offset - candidate_chunk.offset;
-                    let new_free_offset = candidate_chunk.offset + size + candidate_chunk_padding;
-                    let new_free_size = candidate_chunk.size - (candidate_chunk_padding + size);
+                    let new_free_offset = candidate.aligned_offset + size;
+                    let new_free_size =
+                        (candidate_chunk.offset + candidate_chunk.size) - new_free_offset;
 
                     let new_free_chunk = MemoryChunk {
                         block_key: candidate_chunk.block_key,
@@ -792,7 +932,7 @@ impl MemoryPool {
                         offset: new_free_offset,
                         previous: Some(candidate.key),
                         next: candidate_chunk.next,
-                        is_free: true,
+                        chunk_type: ChunkType::Free,
                     };
 
                     let new_free_chunk_key = self.add_chunk(new_free_chunk);
@@ -809,8 +949,8 @@ impl MemoryPool {
                 let candidate_chunk = self.chunks[candidate.key.get()]
                     .as_mut()
                     .expect("can't find chunk in chunk list");
-                candidate_chunk.is_free = false;
-                candidate_chunk.offset = align_up(candidate_chunk.offset, alignment);
+                candidate_chunk.chunk_type = chunk_type;
+                candidate_chunk.offset = candidate.aligned_offset;
                 candidate_chunk.size = size;
 
                 let block = self.blocks[candidate_chunk.block_key.get()]
@@ -826,7 +966,8 @@ impl MemoryPool {
                 };
 
                 let allocation = Allocation {
-                    pool_index: self.pool_index,
+                    memory_type_index: self.memory_type_index,
+                    lifetime,
                     block_key: candidate_chunk.block_key,
                     chunk_key: Some(candidate.key),
                     device_memory: block.device_memory,
@@ -875,7 +1016,7 @@ impl MemoryPool {
             offset: 0,
             previous: None,
             next: None,
-            is_free: true,
+            chunk_type: ChunkType::Free,
         };
 
         let chunk_key = self.add_chunk(chunk);
@@ -891,7 +1032,7 @@ impl MemoryPool {
             let chunk = self.chunks[chunk_key.get()]
                 .as_mut()
                 .ok_or(AllocatorError::CantFindChunk)?;
-            chunk.is_free = true;
+            chunk.chunk_type = ChunkType::Free;
             (chunk.previous, chunk.next, chunk.size)
         };
         self.add_to_free_list(chunk_key, size)?;
@@ -912,7 +1053,8 @@ impl MemoryPool {
             if self.chunks[neighbor_key.get()]
                 .as_ref()
                 .expect("can't find chunk in chunk list")
-                .is_free
+                .chunk_type
+                == ChunkType::Free
             {
                 if neighbor_is_lhs {
                     self.merge_rhs_into_lhs_chunk(neighbor_key, chunk_key)?;
@@ -941,6 +1083,20 @@ impl MemoryPool {
             (chunk.size, chunk.offset, chunk.next)
         };
 
+        let lhs_previous_key = self.chunks[lhs_chunk_key.get()]
+            .as_mut()
+            .expect("can't find chunk in chunk list")
+            .previous;
+
+        let lhs_offset = if let Some(lhs_previous_key) = lhs_previous_key {
+            let lhs_previous = self.chunks[lhs_previous_key.get()]
+                .as_mut()
+                .expect("can't find chunk in chunk list");
+            lhs_previous.offset + lhs_previous.size
+        } else {
+            0
+        };
+
         let lhs_chunk = self.chunks[lhs_chunk_key.get()]
             .as_mut()
             .expect("can't find chunk in chunk list");
@@ -950,7 +1106,8 @@ impl MemoryPool {
         let old_size = lhs_chunk.size;
 
         lhs_chunk.next = rhs_next;
-        lhs_chunk.size = (rhs_offset + rhs_size) - lhs_chunk.offset;
+        lhs_chunk.size = (rhs_offset + rhs_size) - lhs_offset;
+        lhs_chunk.offset = lhs_offset;
 
         let new_size = lhs_chunk.size;
 
@@ -1010,7 +1167,7 @@ struct MemoryChunk {
     offset: vk::DeviceSize,
     previous: Option<NonZeroUsize>,
     next: Option<NonZeroUsize>,
-    is_free: bool,
+    chunk_type: ChunkType,
 }
 
 /// A reserved memory block.
@@ -1053,7 +1210,7 @@ impl MemoryBlock {
         let device_memory = {
             let alloc_info = vk::MemoryAllocateInfoBuilder::new()
                 .allocation_size(size)
-                .memory_type_index(memory_type_index as u32);
+                .memory_type_index(memory_type_index);
 
             unsafe { device.allocate_memory(&alloc_info, None) }
                 .map_err(|_| AllocatorError::OutOfMemory)?
@@ -1095,10 +1252,24 @@ fn align_up(offset: vk::DeviceSize, alignment: vk::DeviceSize) -> vk::DeviceSize
     (offset + (alignment - 1)) & !(alignment - 1)
 }
 
+#[inline]
+fn align_down(offset: vk::DeviceSize, alignment: vk::DeviceSize) -> vk::DeviceSize {
+    offset & !(alignment - 1)
+}
+
+fn is_on_same_page(offset_a: u64, size_a: u64, offset_b: u64, page_size: u64) -> bool {
+    let end_a = offset_a + size_a - 1;
+    let end_page_a = align_down(end_a, page_size);
+    let start_b = offset_b;
+    let start_page_b = align_down(start_b, page_size);
+
+    end_page_a == start_page_b
+}
+
 fn query_driver(
     instance: &erupt::InstanceLoader,
     physical_device: vk::PhysicalDevice,
-) -> (vk::DriverId, bool) {
+) -> (vk::DriverId, bool, u64) {
     let mut vulkan_12_properties = vk::PhysicalDeviceVulkan12Properties::default();
     let physical_device_properties =
         vk::PhysicalDeviceProperties2Builder::new().extend_from(&mut vulkan_12_properties);
@@ -1112,7 +1283,16 @@ fn query_driver(
     let is_integrated =
         physical_device_properties.properties.device_type == vk::PhysicalDeviceType::INTEGRATED_GPU;
 
-    (vulkan_12_properties.driver_id, is_integrated)
+    let buffer_image_granularity = physical_device_properties
+        .properties
+        .limits
+        .buffer_image_granularity;
+
+    (
+        vulkan_12_properties.driver_id,
+        is_integrated,
+        buffer_image_granularity,
+    )
 }
 
 #[inline]
@@ -1156,66 +1336,6 @@ fn calculate_bucket_index(size: vk::DeviceSize) -> u32 {
     }
 }
 
-fn count_unused_ranges(pools: &[Mutex<MemoryPool>]) -> usize {
-    let mut unused_count: usize = 0;
-    pools.iter().for_each(|pool| {
-        collect_start_chunks(pool).iter().for_each(|key| {
-            let mut next_key: NonZeroUsize = *key;
-            let mut previous_size: vk::DeviceSize = 0;
-            let mut previous_offset: vk::DeviceSize = 0;
-            loop {
-                let pool = pool.lock();
-                let chunk = pool.chunks[next_key.get()]
-                    .as_ref()
-                    .expect("can't find chunk in chunk list");
-                if chunk.offset != previous_offset + previous_size {
-                    unused_count += 1;
-                }
-
-                if let Some(key) = chunk.next {
-                    next_key = key
-                } else {
-                    break;
-                }
-
-                previous_size = chunk.size;
-                previous_offset = chunk.offset
-            }
-        });
-    });
-    unused_count
-}
-
-fn count_unused_bytes(pools: &[Mutex<MemoryPool>]) -> vk::DeviceSize {
-    let mut unused_bytes: vk::DeviceSize = 0;
-    pools.iter().for_each(|pool| {
-        collect_start_chunks(pool).iter().for_each(|key| {
-            let mut next_key: NonZeroUsize = *key;
-            let mut previous_size: vk::DeviceSize = 0;
-            let mut previous_offset: vk::DeviceSize = 0;
-            loop {
-                let pool = pool.lock();
-                let chunk = pool.chunks[next_key.get()]
-                    .as_ref()
-                    .expect("can't find chunk in chunk list");
-                if chunk.offset != previous_offset + previous_size {
-                    unused_bytes += chunk.offset - (previous_offset + previous_size);
-                }
-
-                if let Some(key) = chunk.next {
-                    next_key = key
-                } else {
-                    break;
-                }
-
-                previous_size = chunk.size;
-                previous_offset = chunk.offset
-            }
-        });
-    });
-    unused_bytes
-}
-
 #[inline]
 fn collect_start_chunks(pool: &Mutex<MemoryPool>) -> Vec<NonZeroUsize> {
     pool.lock()
