@@ -3,13 +3,17 @@ use rand::Rng;
 use rand_xoshiro::rand_core::SeedableRng;
 use rand_xoshiro::Xoshiro256PlusPlus;
 
-use vk_alloc::{
-    Allocation, AllocationDescriptor, AllocationType, Allocator, AllocatorDescriptor,
-    MemoryLocation,
-};
+use vk_alloc::{Allocation, AllocationDescriptor, Allocator, AllocatorDescriptor, MemoryLocation};
 
 pub mod fixture;
 
+#[derive(Debug, Clone, Copy, Hash, Eq, PartialEq)]
+enum TestLifetime {
+    Static,
+}
+
+impl vk_alloc::Lifetime for TestLifetime {}
+
 #[test]
 fn vulkan_context_creation() {
     fixture::VulkanContext::new(vk::make_api_version(0, 1, 2, 0));
@@ -18,7 +22,7 @@ fn vulkan_context_creation() {
 #[test]
 fn allocator_creation() {
     let ctx = fixture::VulkanContext::new(vk::make_api_version(0, 1, 2, 0));
-    Allocator::new(
+    Allocator::<TestLifetime>::new(
         &ctx.instance,
         ctx.physical_device,
         &AllocatorDescriptor {
@@ -48,8 +52,9 @@ fn allocator_simple_free() {
                     .size(1024)
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -77,7 +82,7 @@ fn allocator_allocation_1024() {
     )
     .unwrap();
 
-    let mut allocations: Vec<Allocation> = (0..1024)
+    let mut allocations: Vec<Allocation<_>> = (0..1024)
         .into_iter()
         .map(|i| {
             let allocation = alloc
@@ -90,8 +95,9 @@ fn allocator_allocation_1024() {
                             .size(1024)
                             .memory_type_bits(u32::MAX)
                             .build(),
-                        allocation_type: AllocationType::Buffer,
+                        lifetime: TestLifetime::Static,
                         is_dedicated: false,
+                        is_optimal: false,
                     },
                 )
                 .unwrap();
@@ -135,7 +141,7 @@ fn allocator_allocation_256() {
     )
     .unwrap();
 
-    let mut allocations: Vec<Allocation> = (0..1024)
+    let mut allocations: Vec<Allocation<_>> = (0..1024)
         .into_iter()
         .map(|i| {
             let allocation = alloc
@@ -148,8 +154,9 @@ fn allocator_allocation_256() {
                             .size(256)
                             .memory_type_bits(u32::MAX)
                             .build(),
-                        allocation_type: AllocationType::Buffer,
+                        lifetime: TestLifetime::Static,
                         is_dedicated: false,
+                        is_optimal: false,
                     },
                 )
                 .unwrap();
@@ -196,7 +203,7 @@ fn allocator_reverse_free() {
     )
     .unwrap();
 
-    let mut allocations: Vec<Allocation> = (0..1024)
+    let mut allocations: Vec<Allocation<_>> = (0..1024)
         .into_iter()
         .map(|i| {
             let allocation = alloc
@@ -209,8 +216,9 @@ fn allocator_reverse_free() {
                             .size(256)
                             .memory_type_bits(u32::MAX)
                             .build(),
-                        allocation_type: AllocationType::Buffer,
+                        lifetime: TestLifetime::Static,
                         is_dedicated: false,
+                        is_optimal: false,
                     },
                 )
                 .unwrap();
@@ -240,9 +248,12 @@ fn allocator_reverse_free() {
 
             assert_eq!(alloc.allocation_count(), 1023 - i);
             assert_eq!(alloc.used_bytes(), 256 * (1023 - i) as vk::DeviceSize);
-            // sic! We free from the front. The padding is part of the next chunk.
-            assert_eq!(alloc.unused_range_count(), 1023 - i);
-            assert_eq!(alloc.unused_bytes(), 768 * (1023 - i) as vk::DeviceSize);
+
+            // the last part not in use is a free chunk!
+            if i < 1023 {
+                assert_eq!(alloc.unused_range_count(), 1022 - i);
+                assert_eq!(alloc.unused_bytes(), 768 * (1022 - i) as vk::DeviceSize);
+            }
         });
 
     assert_eq!(alloc.allocation_count(), 0);
@@ -263,7 +274,7 @@ fn allocator_free_every_second_time() {
     )
     .unwrap();
 
-    let allocations: Vec<Allocation> = (0..1024)
+    let allocations: Vec<Allocation<_>> = (0..1024)
         .into_iter()
         .map(|_| {
             let allocation = alloc
@@ -276,8 +287,9 @@ fn allocator_free_every_second_time() {
                             .size(1024)
                             .memory_type_bits(u32::MAX)
                             .build(),
-                        allocation_type: AllocationType::Buffer,
+                        lifetime: TestLifetime::Static,
                         is_dedicated: false,
+                        is_optimal: false,
                     },
                 )
                 .unwrap();
@@ -285,13 +297,13 @@ fn allocator_free_every_second_time() {
         })
         .collect();
 
-    let mut odd: Vec<Allocation> = allocations
+    let mut odd: Vec<Allocation<_>> = allocations
         .iter()
         .enumerate()
         .filter(|(index, _)| index % 2 == 0)
         .map(|(_, allocation)| allocation.clone())
         .collect();
-    let mut even: Vec<Allocation> = allocations
+    let mut even: Vec<Allocation<_>> = allocations
         .iter()
         .enumerate()
         .filter(|(index, _)| index % 2 == 1)
@@ -334,8 +346,9 @@ fn allocator_allocation_dedicated() {
                     .size(10 * 1024 * 1024) // 10 MiB
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -377,8 +390,9 @@ fn allocator_properly_merge_free_entries() {
                     .size(256)
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -392,8 +406,9 @@ fn allocator_properly_merge_free_entries() {
                     .size(256)
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -407,8 +422,9 @@ fn allocator_properly_merge_free_entries() {
                     .size(256)
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -422,8 +438,9 @@ fn allocator_properly_merge_free_entries() {
                     .size(256)
                     .memory_type_bits(u32::MAX)
                     .build(),
-                allocation_type: AllocationType::Buffer,
+                lifetime: TestLifetime::Static,
                 is_dedicated: false,
+                is_optimal: false,
             },
         )
         .unwrap();
@@ -446,7 +463,7 @@ fn allocator_fuzzy() {
     )
     .unwrap();
 
-    let mut allocations: Vec<(u8, Allocation)> = Vec::with_capacity(10_0000);
+    let mut allocations: Vec<(u8, Allocation<_>)> = Vec::with_capacity(10_0000);
     let mut rng = Xoshiro256PlusPlus::seed_from_u64(0);
 
     for i in 0..10_000 {
@@ -466,8 +483,9 @@ fn allocator_fuzzy() {
                             .size(size as u64)
                             .memory_type_bits(u32::MAX)
                             .build(),
-                        allocation_type: AllocationType::Buffer,
+                        lifetime: TestLifetime::Static,
                         is_dedicated: false,
+                        is_optimal: false,
                     },
                 )
                 .unwrap();
@@ -492,3 +510,78 @@ fn allocator_fuzzy() {
     }
     alloc.cleanup(&ctx.logical_device);
 }
+
+#[test]
+fn allocator_granularity() {
+    let ctx = fixture::VulkanContext::new(vk::make_api_version(0, 1, 2, 0));
+    let mut alloc = Allocator::new(
+        &ctx.instance,
+        ctx.physical_device,
+        &AllocatorDescriptor { block_size: 20 }, // 1 MiB
+    )
+    .unwrap();
+
+    let allocation1 = alloc
+        .allocate(
+            &ctx.logical_device,
+            &AllocationDescriptor {
+                location: MemoryLocation::GpuOnly,
+                requirements: vk::MemoryRequirementsBuilder::new()
+                    .alignment(256)
+                    .size(512)
+                    .memory_type_bits(u32::MAX)
+                    .build(),
+                lifetime: TestLifetime::Static,
+                is_dedicated: false,
+                is_optimal: false,
+            },
+        )
+        .unwrap();
+
+    assert_eq!(allocation1.size, 512);
+    assert_eq!(allocation1.offset, 0);
+
+    let allocation2 = alloc
+        .allocate(
+            &ctx.logical_device,
+            &AllocationDescriptor {
+                location: MemoryLocation::GpuOnly,
+                requirements: vk::MemoryRequirementsBuilder::new()
+                    .alignment(256)
+                    .size(1024)
+                    .memory_type_bits(u32::MAX)
+                    .build(),
+                lifetime: TestLifetime::Static,
+                is_dedicated: false,
+                is_optimal: true,
+            },
+        )
+        .unwrap();
+
+    assert_eq!(allocation2.size, 1024);
+    assert_eq!(allocation2.offset, 1024);
+
+    alloc.deallocate(&ctx.logical_device, &allocation2).unwrap();
+
+    let allocation3 = alloc
+        .allocate(
+            &ctx.logical_device,
+            &AllocationDescriptor {
+                location: MemoryLocation::GpuOnly,
+                requirements: vk::MemoryRequirementsBuilder::new()
+                    .alignment(256)
+                    .size(1024)
+                    .memory_type_bits(u32::MAX)
+                    .build(),
+                lifetime: TestLifetime::Static,
+                is_dedicated: false,
+                is_optimal: false,
+            },
+        )
+        .unwrap();
+
+    assert_eq!(allocation3.size, 1024);
+    assert_eq!(allocation3.offset, 512);
+
+    alloc.cleanup(&ctx.logical_device);
+}
